"""Visualization tools."""
import numpy as np
import matplotlib.pyplot as plt
import pytransform3d.visualizer as pv
import open3d as o3d


class Mesh(pv.Artist):
    """A triangular mesh.

    Parameters
    ----------
    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh.

    c : array-like, shape (3,), optional (default: None)
        Color
    """
    def __init__(self, mesh2origin, vertices, triangles, c=None):
        self.mesh = o3d.geometry.TriangleMesh()
        self.mesh.vertices = o3d.utility.Vector3dVector(vertices)
        self.mesh.triangles = o3d.utility.Vector3iVector(triangles)
        self.mesh.compute_vertex_normals()
        if c is not None:
            self.mesh.paint_uniform_color(c)
        self.mesh2origin = None
        self.set_data(mesh2origin)

    def set_data(self, mesh2origin):
        """Update data.

        Parameters
        ----------
        mesh2origin : array-like, shape (4, 4)
            Pose of the mesh.
        """
        previous_mesh2origin = self.mesh2origin
        if previous_mesh2origin is None:
            previous_mesh2origin = np.eye(4)
        self.mesh2origin = mesh2origin

        self.mesh.transform(np.linalg.inv(previous_mesh2origin))
        self.mesh.transform(self.mesh2origin)

    @property
    def geometries(self):
        """Expose geometries.

        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.mesh]


class TetraMesh(pv.Artist):
    """A tetrahedral mesh.

    Parameters
    ----------
    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    c : array-like, shape (3,), optional (default: None)
        Color
    """
    def __init__(self, mesh2origin, vertices, tetrahedra, c=None):
        self.mesh = o3d.geometry.TetraMesh(
            o3d.utility.Vector3dVector(vertices),
            o3d.utility.Vector4iVector(tetrahedra))
        if c is not None:
            self.mesh.paint_uniform_color(c)
        self.mesh2origin = None
        self.set_data(mesh2origin)

    def set_data(self, mesh2origin):
        """Update data.

        Parameters
        ----------
        mesh2origin : array-like, shape (4, 4)
            Pose of the mesh.
        """
        previous_mesh2origin = self.mesh2origin
        if previous_mesh2origin is None:
            previous_mesh2origin = np.eye(4)
        self.mesh2origin = mesh2origin

        self.mesh.transform(np.linalg.inv(previous_mesh2origin))
        self.mesh.transform(self.mesh2origin)

    @property
    def geometries(self):
        """Expose geometries.

        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.mesh]


class Tetrahedron(Mesh):  # pragma: no cover
    """Tetrahedron.

    Parameters
    ----------
    tetrahedron_points : array, shape (4, 3)
        Points of the tetrahedron.

    c : array-like, shape (3,), optional (default: None)
        Color
    """
    def __init__(self, tetrahedron_points, c=None):
        mesh2origin = np.eye(4)
        triangles = np.array([[0, 1, 2], [1, 3, 2], [3, 0, 2], [0, 3, 1]], dtype=int)
        super(Tetrahedron, self).__init__(mesh2origin, tetrahedron_points, triangles, c)


class Ellipse(pv.Artist):  # pragma: no cover
    """An ellipse.

    Parameters
    ----------
    center : array, shape (3,)
        Center of ellipse.

    axes : array, shape (2, 3)
        Axes of ellipse.

    radii : array, shape (2,)
        Radii of ellipse.

    c : array-like, shape (3,), optional (default: None)
        Color(s)
    """
    def __init__(self, center, axes, radii, c=None):
        self.mesh = o3d.geometry.TriangleMesh()
        ellipse2origin = np.eye(4)
        ellipse2origin[:3, :2] = axes.T
        ellipse2origin[:3, 2] = np.cross(axes[0], axes[1])
        ellipse2origin[:3, 3] = center
        ellipse = np.array([
            np.hstack(((radii * np.array([np.cos(angle), np.sin(angle)])), (0.0,)))
            for angle in np.linspace(0, 2 * np.pi, 20)])
        vertices = np.vstack((np.zeros((1, 3)), ellipse))
        triangles = np.vstack((
            np.array([[i, j, 0] for i, j in zip(range(1, len(vertices) - 1),
                                                range(2, len(vertices)))]),
            np.array([[0, j, i] for i, j in zip(range(1, len(vertices) - 1),
                                                range(2, len(vertices)))]),
        ))
        self.mesh.vertices = o3d.utility.Vector3dVector(vertices)
        self.mesh.triangles = o3d.utility.Vector3iVector(triangles)
        self.mesh.compute_vertex_normals()
        if c is not None:
            self.mesh.paint_uniform_color(c)
        self.ellipse2origin = None
        self.set_data(ellipse2origin)

    def set_data(self, ellipse2origin):
        """Update data.

        Parameters
        ----------
        ellipse2origin : array-like, shape (4, 4)
            Pose of the ellipse.
        """
        previous_ellipse2origin = self.ellipse2origin
        if previous_ellipse2origin is None:
            previous_ellipse2origin = np.eye(4)
        self.ellipse2origin = ellipse2origin

        self.mesh.transform(np.linalg.inv(previous_ellipse2origin))
        self.mesh.transform(self.ellipse2origin)

    @property
    def geometries(self):
        """Expose geometries.

        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.mesh]


class ContactSurface(pv.Artist):  # pragma: no cover
    """A pressure field.

    Parameters
    ----------
    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    contact_vertices : list
        Each entry is an array of vertices of a contact polygon.

    contact_triangles : list
        Each entry is an array of trangles of a contact polygon.

    pressures : array, shape (n_contact_polygons)
        Pressure values for contact polygons.
    """
    def __init__(self, mesh2origin, contact_vertices, contact_triangles, pressures):
        self.cmap = plt.get_cmap("plasma")
        self.mesh = o3d.geometry.TriangleMesh()
        self.set_data(mesh2origin, contact_vertices, contact_triangles, pressures)

    def set_data(self, mesh2origin, contact_vertices, contact_triangles, pressures):
        """Update contact surface.

        Parameters
        ----------
        mesh2origin : array, shape (4, 4)
            Pose of the mesh.

        contact_vertices : list
            Each entry is an array of vertices of a contact polygon.

        contact_triangles : list
            Each entry is an array of trangles of a contact polygon.

        pressures : array, shape (n_contact_polygons)
            Pressure values for contact polygons.
        """
        if len(contact_vertices) > 0:
            vertices = np.vstack(contact_vertices)
            triangles = []
            n_vertices = 0
            pressures_per_face = []
            for i in range(len(contact_triangles)):
                triangles.extend(contact_triangles[i] + n_vertices)
                n_vertices_i = len(contact_vertices[i])
                n_vertices += n_vertices_i
                pressures_per_face.extend([pressures[i]] * n_vertices_i)
            triangles = np.vstack(triangles)
            triangles = np.vstack((triangles, triangles[:, ::-1]))
            max_pressure = max(pressures)
            if max_pressure <= 0.0:
                max_pressure = 1.0
            colors = np.vstack([self.cmap(pressure_per_face / max_pressure)[:3]
                                for pressure_per_face in pressures_per_face])
        else:
            vertices = np.empty((0, 3), dtype=float)
            triangles = np.empty((0, 4), dtype=int)
            colors = np.empty((0, 3), dtype=float)
        self.mesh.vertices = o3d.utility.Vector3dVector(vertices)
        self.mesh.triangles = o3d.utility.Vector3iVector(triangles)
        self.mesh.vertex_colors = o3d.utility.Vector3dVector(colors)
        self.mesh.transform(mesh2origin)

    @property
    def geometries(self):
        """Expose geometries.

        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.mesh]


class RigidBodyTetrahedralMesh(pv.Artist):  # pragma: no cover
    """Rigid body represented by a tetrahedral mesh.

    Parameters
    ----------
    body2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    c : array-like, shape (3,), optional (default: None)
        Color(s)
    """
    def __init__(self, body2origin, vertices, tetrahedra, c=None):
        self.mesh = o3d.geometry.TetraMesh()
        self.set_data(body2origin, vertices, tetrahedra, c)

    def set_data(self, body2origin, vertices, tetrahedra, c=None):
        """Update rigid body.

        Parameters
        ----------
        body2origin : array, shape (4, 4)
            Pose of the mesh.

        vertices : array, shape (n_vertices, 3)
            Vertices of the mesh.

        tetrahedra : array, shape (n_tetrahedra, 4)
            Indices of vertices that form tetrahedra of the mesh.

        c : array-like, shape (3,), optional (default: None)
            Color(s)
        """
        self.mesh.vertices = o3d.utility.Vector3dVector(vertices)
        self.mesh.tetras = o3d.utility.Vector4iVector(tetrahedra)
        self.mesh.transform(body2origin)
        if c is not None:
            self.mesh.paint_uniform_color(c)

    @property
    def geometries(self):
        """Expose geometries.

        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.mesh]


class ColoredVertices(pv.Artist):  # pragma: no cover
    """Vertices of a mesh without connections with color information.

    Parameters
    ----------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    colors : array-like, shape (n_vertices, 3)
        Colors
    """
    def __init__(self, vertices, colors):
        self.vertices = o3d.geometry.PointCloud(
            o3d.utility.Vector3dVector(vertices))
        self.vertices.colors = o3d.utility.Vector3dVector(colors)

    @property
    def geometries(self):
        """Expose geometries.

        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.vertices]



from distance3d.broad_phase import BoundingVolumeHierarchy
from pytransform3d import urdf
from distance3d.hydroelastic_contact import RigidBody
import numpy as np


class HydroelasticBoundingVolumeHierarchy(BoundingVolumeHierarchy):
    """Hydroelastic BVH for broad phase collision detection.

    This BVH works the same but uses the hydroelastic RigidBody as the
    collider object.

    Parameters
    ----------
    tm : pytransform3d.transform_manager.TransformManager
        Transform manager that stores the transformations.

    base_frame : str
        Name of the base frame in which colliders are represented.

    base_frame2origin : array, shape (4, 4), optional (default: np.eye(4))
        The position of the base_frame to origin.

    Attributes
    ----------
    aabbtree_ : AabbTree
        Tree of axis-aligned bounding boxes.

    colliders_ : dict
        Maps frames of collision objects to colliders.

    self_collision_whitelists_ : dict
        Whitelists for self-collision detection in case this BVH represents
        a robot.
    """

    def __init__(self, tm, base_frame, base_frame2origin=np.eye(4)):
        super().__init__(tm, base_frame, base_frame2origin)

    def _make_collider(self, tm, obj, make_artists):
        """Creates a collider from a URDF object.

        Parameters
        ----------
        tm : pytransform3d.urdf.UrdfTransformManager
            Transform manager that has colliders.

        obj: pytransform3d.urdf.Geometry
            The urdf object.

        make_artists : bool, optional (default: False)
            Create artist for visualization for each collision object.

        Returns
        -------
        collider : ConvexCollider
            The corresponding collider.
        """
        A2B = tm.get_transform(obj.frame, "origin")

        if isinstance(obj, urdf.Sphere):
            collider = RigidBody.make_sphere(A2B[:3, 3], obj.radius, 2)
        elif isinstance(obj, urdf.Box):
            collider = RigidBody.make_box(A2B, obj.size)
        elif isinstance(obj, urdf.Cylinder):
            collider = RigidBody.make_cylinder(
                A2B, obj.radius, obj.length, resolution_hint=0.01)
        else:  # pragma: no cover
            assert isinstance(obj, urdf.Mesh)
            raise NotImplementedError(
                "Arbitrary mesh conversion is not implemented!")

        if make_artists:
            collider.make_artist()
        return collider

    # Override to change the docstring.
    def aabb_overlapping_colliders(self, collider, whitelist=()):
        """Get colliders with an overlapping AABB.

        This function performs broad phase collision detection with a bounding
        volume hierarchy, where the bounding volumes are axis-aligned bounding
        boxes.

        Parameters
        ----------
        collider : RigidBody
            RigidBody.

        whitelist : sequence
            Names of frames to which collisions are allowed.

        Returns
        -------
        colliders : dict
            Maps frame names to colliders with overlapping AABB.
        """
        return super().aabb_overlapping_colliders(collider, whitelist)



import numpy as np
import numba

INDEX_NONE = -1
PARENT_INDEX = 0
LEFT_INDEX = 1
RIGHT_INDEX = 2
TYPE_INDEX = 3
TYPE_NONE = -1
TYPE_LEAF = 1
TYPE_BRANCH = 2


class AabbTree:
    """AABB Tree for broad phase collision detection."""
    def __init__(self):
        self.root = INDEX_NONE
        self.filled_len = 0
        self.nodes = np.empty((0, 4), dtype=int)
        self.aabbs = np.empty((0, 3, 2))
        self.external_data_list = []

    def insert_aabbs(self, aabbs, external_data_list=None,
                     pre_insertion_methode="none"):
        """Insert aabbs in tree

        Parameters
        ----------
        aabbs : array, shape (n, 3, 2)
            An array containing a list of aabbs of the tree.

        external_data_list : list, optional (default: None)
            Array of data associated with the aabbs.

        pre_insertion_methode : str, optional (default: "none")
            The operation that is performed on the aabbs before tree creation.
            Use "sort" for a cleaner tree with slightly longer creation times.
            Use "shuffle" for a faster creation but with some non-optimal placement in the tree.
        """
        if len(aabbs) == 0:
            return

        assert external_data_list is None or len(external_data_list) == len(aabbs)

        old_filled_len = self.filled_len
        self.filled_len += len(aabbs)

        # The maximum number of possible nodes a tree can have is 2n
        new_nodes = np.full([2 * (self.filled_len - len(self.nodes)), 4], INDEX_NONE)
        self.nodes = np.append(self.nodes, new_nodes, axis=0)

        self.aabbs = np.append(self.aabbs, aabbs, axis=0)
        new_aabbs = np.zeros([len(self.nodes) - len(self.aabbs), 3, 2])
        self.aabbs = np.append(self.aabbs, new_aabbs, axis=0)

        if external_data_list is not None:
            self.external_data_list += external_data_list
        empty_external_data = [None] * (len(self.nodes) - len(self.external_data_list))
        self.external_data_list += empty_external_data

        insert_order = np.array(range(old_filled_len, self.filled_len))
        if pre_insertion_methode == "sort":
            insert_order = _sort_aabbs(aabbs[old_filled_len:len(self.nodes) - self.filled_len])
        elif pre_insertion_methode == "shuffle":
            np.random.shuffle(insert_order)

        self.root, self.nodes, self.aabbs, self.filled_len \
            = insert_aabbs(self.root, self.nodes, self.aabbs, self.filled_len, insert_order)

        self.nodes = self.nodes[:self.filled_len]
        self.aabbs = self.aabbs[:self.filled_len]
        self.external_data_list = self.external_data_list[:self.filled_len]

    def insert_aabb(self, aabb, external_data=None):
        """Insert single aabb in tree

        Parameters
        ----------
        aabb : array, shape (3, 2)
            The aabb that will be inserted into the tree.

        external_data : Any
            Data associated with the aabb.
        """

        self.insert_aabbs([aabb], [external_data], pre_insertion_methode="none")

    def __str__(self):  # pragma: no cover
        lines, *_ = print_aabb_tree_recursive(self.root, self.nodes)
        return '\r\n' + '\r\n'.join(lines)

    def overlaps_aabb_tree(self, other):
        """Check overlapping with another tree.

        Parameters
        ----------
        other : AabbTree
            The other Tree for overlap testing.

        Returns
        -------
        is_overlapping : bool
            True if there is an overlap in the two trees.

        overlap_tetrahedron1 : array, shape (n)
            The indexes of the overlapping tetrahedron in this tree.

        overlap_tetrahedron2 : array, shape (n)
            The indexes of the overlapping tetrahedron in the other tree.

        overlap_pairs : array, shape (n, 2)
            An array of all overlapping pairs.
        """
        overlap_tetrahedron1, overlap_tetrahedron2, overlap_pairs = query_overlap_of_other_tree(
            self.root, self.nodes, self.aabbs, other.root, other.nodes, other.aabbs)

        is_overlapping = len(overlap_pairs) > 0
        return is_overlapping, np.unique(overlap_tetrahedron1), np.unique(overlap_tetrahedron2), overlap_pairs

    def overlaps_aabb(self, aabb):
        """Check overlapping with an aabb.

        Parameters
        ----------
        aabb : array, shape (3, 2)
            The aabb that is checked.

        Returns
        -------
        is_overlapping : bool
            True if there is an overlap in the two trees.

        overlaps : array, shape (n, 2)
            An array of all overlapping pairs.
        """
        overlaps = query_overlap(aabb, self.root, self.nodes, self.aabbs)
        return len(overlaps) > 0, overlaps

    def get_root_aabb(self):
        """Return the aabb of the entire tree.

        Returns
        -------
        aabb : array, shape (3, 2)
            The aabb of the root node.
        """
        return self.aabbs[self.root]


@numba.njit(cache=True)
def insert_aabbs(root, nodes, aabbs, filled_len, insert_order):
    """Inserts aabbs into the tree defined in root and nodes.

    Parameters
    ----------
    root : int
        The index of the tree root node.

    nodes : array, shape (n, 4)
        The index links of the tree structure.
        The indices of the 1-axis correspond to the aabb indices.
        The first entry in of the 2-axis is the index of the parent.
        The second entry in of the 2-axis is the index of the left child.
        The third entry in of the 2-axis is the index of the right child.
        The forth entry in of the 2-axis is 1 if leaf and 2 if branch node.

    aabbs : array, shape (3, 2)
        The aabbs that are inserted.

    filled_len : int
        The length to which the nodes and aabbs are filled.

    insert_order : array, shape (filled_len)
        A list of indexes describing the insert order into the tree.

    Returns
    -------
    root : int
        The index of the tree root node.

    nodes : array, shape (n, 4)
        The index links of the tree structure.

    aabbs : array, shape (3, 2)
        The aabbs of the tree.

    filled_len : int
        The length to which the nodes and aabbs are filled.
    """
    for i in insert_order:
        root, nodes, aabbs, filled_len = insert_leaf(root, i, nodes, aabbs, filled_len)

    return root, nodes, aabbs, filled_len


@numba.njit(cache=True)
def insert_leaf(root_node_index, leaf_node_index, nodes, aabbs, filled_len):
    """
    Inserts a new leaf into the tree.
    """
    # From https://github.com/JamesRandall/SimpleVoxelEngine/blob/master/voxelEngine/src/AABBTree.cpp
    # The Node that's going to be added.
    # A node in this system consists of three variables. An int as index.
    # The node array with containing [parent_index, left_child_index, right_child_index, type]

    nodes[leaf_node_index, TYPE_INDEX] = TYPE_LEAF

    # If there is no root make new leaf root.
    if root_node_index == INDEX_NONE:
        return leaf_node_index, nodes, aabbs, filled_len

    # Traverse the tree down till you find a leaf.
    tree_node_index = root_node_index
    while nodes[tree_node_index, TYPE_INDEX] == TYPE_BRANCH:

        # Getting nodes from arrays
        left_node_index = nodes[tree_node_index, LEFT_INDEX]
        right_node_index = nodes[tree_node_index, RIGHT_INDEX]

        # Whether the left or right child is traverse down depends on the cost of increasing the size of the aabbs.
        cost_new_parent = _aabb_volume(
            _merge_aabb(aabbs[leaf_node_index], aabbs[tree_node_index]))

        cost_left = _aabb_volume(
            _merge_aabb(aabbs[leaf_node_index], aabbs[left_node_index]))

        cost_right = _aabb_volume(
            _merge_aabb(aabbs[leaf_node_index], aabbs[right_node_index]))

        # This doesn't seem to happen ever. So need to check if it does
        assert not (cost_left > cost_new_parent and cost_right > cost_new_parent)

        # otherwise go down the cheaper child.
        if cost_left < cost_right:
            tree_node_index = left_node_index
        else:
            tree_node_index = right_node_index

    # Setting up sibling node
    sibling_index = tree_node_index

    # Inserting new leaf into tree
    old_parent_index = nodes[sibling_index, PARENT_INDEX]

    # Adding Parent
    new_parent_index = filled_len
    filled_len += 1

    nodes[new_parent_index, PARENT_INDEX] = old_parent_index
    nodes[new_parent_index, LEFT_INDEX] = sibling_index
    nodes[new_parent_index, RIGHT_INDEX] = leaf_node_index
    nodes[new_parent_index, TYPE_INDEX] = TYPE_BRANCH
    aabbs[new_parent_index] = _merge_aabb(aabbs[leaf_node_index], aabbs[sibling_index])

    # Setting Parent in children
    nodes[leaf_node_index, PARENT_INDEX] = new_parent_index
    nodes[sibling_index, PARENT_INDEX] = new_parent_index

    if old_parent_index == INDEX_NONE:
        root_node_index = new_parent_index
    else:
        if nodes[old_parent_index, LEFT_INDEX] == sibling_index:
            nodes[old_parent_index, LEFT_INDEX] = new_parent_index
        else:
            nodes[old_parent_index, RIGHT_INDEX] = new_parent_index

    # Set the index to the parent for upward traversal.
    tree_node_index = nodes[leaf_node_index, PARENT_INDEX]
    aabbs = fix_upward_tree(tree_node_index, nodes, aabbs)

    return root_node_index, nodes, aabbs, filled_len


@numba.njit(cache=True)
def fix_upward_tree(tree_node_index, nodes, aabbs):
    """Fixes the aabbs of the parent branches by setting them to the merge of the children aabbs."""

    # Go the tree back up while fixing the aabbs.
    while tree_node_index != INDEX_NONE:
        tree_node = nodes[tree_node_index]

        # Every branch in the traversal should not have any None fields.
        assert tree_node[LEFT_INDEX] != INDEX_NONE and tree_node[RIGHT_INDEX] != INDEX_NONE

        aabbs[tree_node_index] = _merge_aabb(aabbs[tree_node[LEFT_INDEX]], aabbs[tree_node[RIGHT_INDEX]])

        # Moving on parent up
        tree_node_index = tree_node[PARENT_INDEX]

    return aabbs


@numba.njit(cache=True)
def query_overlap_of_other_tree(root1, nodes1, aabbs1, root2, nodes2, aabbs2):  # pragma: no cover
    """Queries the overlapping aabbs by traversing the trees."""

    broad_tetrahedra1 = []
    broad_tetrahedra2 = []
    stack = [root2]

    while len(stack) != 0:

        node_index = stack[-1]
        stack = stack[:-1]

        node_aabb = aabbs2[node_index]
        if nodes2[node_index, TYPE_INDEX] == TYPE_BRANCH and \
                len(query_overlap(node_aabb, root1, nodes1, aabbs1, break_at_first_leaf=True)) >= 1:
            stack.extend([nodes2[node_index, 1], nodes2[node_index, 2]])

        elif nodes2[node_index, TYPE_INDEX] == TYPE_LEAF:
            overlaps = query_overlap(node_aabb, root1, nodes1, aabbs1)
            broad_tetrahedra1.extend(overlaps)
            broad_tetrahedra2.extend([node_index] * len(overlaps))

    broad_pairs = list(zip(broad_tetrahedra1, broad_tetrahedra2))

    return np.array(broad_tetrahedra1), np.array(broad_tetrahedra2), broad_pairs


@numba.njit(cache=True)
def query_overlap(test_aabb, root_node_index, nodes, aabbs, break_at_first_leaf=False):
    """Queries the overlapping aabbs by traversing the tree."""
    overlaps = []
    stack = [root_node_index]

    while len(stack) != 0:

        node_index = stack[-1]
        stack = stack[:-1]

        node_aabb = aabbs[node_index]
        if aabb_overlap(node_aabb, test_aabb):

            if nodes[node_index, TYPE_INDEX] == TYPE_LEAF:
                overlaps.extend([node_index])

                if break_at_first_leaf:
                    break
            else:
                stack.extend([nodes[node_index, 1], nodes[node_index, 2]])

    return np.array(overlaps)


def print_aabb_tree_recursive(node_index, nodes):  # pragma: no cover
    """Returns list of strings, width, height, and horizontal coordinate of the root."""
    # From https://stackoverflow.com/questions/34012886/print-binary-tree-level-by-level-in-python

    # No child.
    line = '%s' % node_index
    width = len(line)

    if nodes[node_index, LEFT_INDEX] == INDEX_NONE and nodes[node_index, RIGHT_INDEX] == INDEX_NONE:
        line = '%s' % node_index
        height = 1
        middle = width // 2
        return [line], width, height, middle

    # Only left child.
    if nodes[node_index, RIGHT_INDEX] == INDEX_NONE:
        lines, n, p, x = print_aabb_tree_recursive(nodes[node_index, LEFT_INDEX], nodes)

        first_line = (x + 1) * ' ' + (n - x - 1) * '_' + line
        second_line = x * ' ' + '/' + (n - x - 1 + width) * ' '
        shifted_lines = [line + width * ' ' for line in lines]
        return [first_line, second_line] + shifted_lines, n + width, p + 2, n + width // 2

    # Only right child.
    if nodes[node_index, LEFT_INDEX] == INDEX_NONE:
        lines, n, p, x = print_aabb_tree_recursive(nodes[node_index, RIGHT_INDEX], nodes)

        first_line = line + x * '_' + (n - x) * ' '
        second_line = (width + x) * ' ' + '\\' + (n - x - 1) * ' '
        shifted_lines = [width * ' ' + line for line in lines]
        return [first_line, second_line] + shifted_lines, n + width, p + 2, width // 2

    # Two children.
    left, n, p, x = print_aabb_tree_recursive(nodes[node_index, LEFT_INDEX], nodes)
    right, m, q, y = print_aabb_tree_recursive(nodes[node_index, RIGHT_INDEX], nodes)

    first_line = (x + 1) * ' ' + (n - x - 1) * '_' + line + y * '_' + (m - y) * ' '
    second_line = x * ' ' + '/' + (n - x - 1 + width + y) * ' ' + '\\' + (m - y - 1) * ' '
    if p < q:
        left += [n * ' '] * (q - p)
    elif q < p:
        right += [m * ' '] * (p - q)
    zipped_lines = zip(left, right)
    lines = [first_line, second_line] + [a + width * ' ' + b for a, b in zipped_lines]
    return lines, n + m + width, max(p, q) + 2, n + width // 2


@numba.njit(cache=True)
def all_aabbs_overlap(aabbs1, aabbs2):
    """Creates result lists of all the overlapping aabbs (brute force).

    Parameters
    ----------
    aabbs1 : array, shape (n, 3, 2)
        The AABBs of the first object.

    aabbs2 : array, shape (n, 3, 2)
        The AABBs of the second object.

    Returns
    -------
    broad_tetrahedra1 : array, shape (n)
        Array of all the overlapping aabb indices of aabbs1

    broad_tetrahedra2 : array, shape (n)
        Array of all the overlapping aabb indices of aabbs2

    broad_pairs : array, shape(n, 2)
        A list of a index pairs of the all overlaps.
    """

    indices1 = []
    indices2 = []
    broad_pairs = []
    for i in range(len(aabbs1)):
        for j in range(len(aabbs2)):
            if aabb_overlap(aabbs1[i], aabbs2[j]):
                indices1.append(i)
                indices2.append(j)
                broad_pairs.append((i, j))
    broad_tetrahedra1 = np.unique(np.array(indices1, dtype=np.dtype("int")))
    broad_tetrahedra2 = np.unique(np.array(indices2, dtype=np.dtype("int")))
    return broad_tetrahedra1, broad_tetrahedra2, broad_pairs


@numba.njit(cache=True)
def aabb_overlap(aabb1, aabb2):
    """Returns true if aabb1 and aabb2 overlap.

    Parameters
    ----------
    aabb1 : array, shape (3, 2)
        The AABB of the first object.

    aabb2 : array, shape (3, 2)
        The AABB of the second object.

    Returns
    -------
    overlap : bool
        Do both AABBs overlap?
    """
    return aabb1[0, 0] <= aabb2[0, 1] and aabb1[0, 1] >= aabb2[0, 0] \
           and aabb1[1, 0] <= aabb2[1, 1] and aabb1[1, 1] >= aabb2[1, 0] \
           and aabb1[2, 0] <= aabb2[2, 1] and aabb1[2, 1] >= aabb2[2, 0]


@numba.njit(cache=True)
def _sort_aabbs(aabbs):
    """Returns a spatially sorted aabb list."""
    return aabbs[:, 0, 0].argsort()


@numba.njit(cache=True)
def _merge_aabb(aabb1, aabb2):
    """Returns the smallest aabb that contains aabb1 and aabb2."""
    return np.array(
        [[min(aabb1[0, 0], aabb2[0, 0]), max(aabb1[0, 1], aabb2[0, 1])],
         [min(aabb1[1, 0], aabb2[1, 0]), max(aabb1[1, 1], aabb2[1, 1])],
         [min(aabb1[2, 0], aabb2[2, 0]), max(aabb1[2, 1], aabb2[2, 1])]]
    )


@numba.njit(cache=True)
def _aabb_volume(aabb):
    """Returns the volume of the aabb."""
    return _aabb_x_size(aabb) * _aabb_y_size(aabb) * _aabb_z_size(aabb)


@numba.njit(cache=True)
def _aabb_x_size(aabb):
    """Returns the size of the aabb along the x-axsis."""
    return aabb[0, 1] - aabb[0, 0]


@numba.njit(cache=True)
def _aabb_y_size(aabb):
    """Returns the size of the aabb along the y-axsis."""
    return aabb[1, 1] - aabb[1, 0]


@numba.njit(cache=True)
def _aabb_z_size(aabb):
    """Returns the size of the aabb along the Z-axsis."""
    return aabb[2, 1] - aabb[2, 0]



import numpy as np
from ..utils import transform_points, transform_directions


class ContactSurface:
    """Contact surface of tetrahedral meshes.

    Parameters
    ----------
    frame2world : array, shape (4, 4)
        Transforms vertices, points, and directions to world frame.

    intersection : bool
        Do both tetrahedral meshes intersect?

    contact_planes : array, shape (n_intersections, 4)
        Contact planes of intersection pairs in Hesse normal form.

    contact_polygons : list
        Contact polygons of intersection pairs.

    intersecting_tetrahedra1 : list
        Intersecting tetrahedron indices of first mesh.

    intersecting_tetrahedra2 : list
        Intersecting tetrahedron indices of second mesh.
    """
    def __init__(
            self, frame2world, intersection, contact_planes, contact_polygons,
            intersecting_tetrahedra1, intersecting_tetrahedra2):
        self.frame2world = frame2world
        self.intersection = intersection
        self.contact_planes = contact_planes
        self.contact_polygons = contact_polygons
        self.intersecting_tetrahedra1 = intersecting_tetrahedra1
        self.intersecting_tetrahedra2 = intersecting_tetrahedra2

        self.contact_areas = None
        self.contact_coms = None
        self.contact_forces = None
        self.contact_polygon_triangles = None

    def add_polygon_info(self, contact_areas, contact_coms, contact_forces,
                         contact_polygon_triangles):
        """Add more info on contact polygons.

        Parameters
        ----------
        contact_areas : array, shape (n_intersections,)
            Areas of contact polygons.

        contact_coms : array, shape (n_intersections, 3)
            Center of mass of contact polygons.

        contact_forces : array, shape (n_intersections, 3)
            Contact forces of contact polygons.

        contact_polygon_triangles : list
            Vertex indices of triangles of contact polygons.
        """
        self.contact_areas = contact_areas
        self.contact_coms = contact_coms
        self.contact_forces = contact_forces
        self.contact_polygon_triangles = contact_polygon_triangles

    @property
    def pressures(self):
        contact_areas = np.copy(self.contact_areas)
        contact_areas[contact_areas == 0.0] = 1.0
        return np.linalg.norm(self.contact_forces, axis=1) / contact_areas

    def make_details(self, tetrahedra_points1, tetrahedra_points2):
        """Summarize details of contact points in world frame.

        Parameters
        ----------
        tetrahedra_points1 : array, shape (n_tetrahedra1, 4, 3)
            Points that form the first tetrahedra.

        tetrahedra_points2 : array, shape (n_tetrahedra2, 4, 3)
            Points that form the second tetrahedra.

        Returns
        -------
        details : dict
            Detailed description of contact surface and pressure field.
        """
        if not self.intersection:
            return {}

        intersecting_tetrahedra1, intersecting_tetrahedra2 = self._transform_to_world(
            self.frame2world, tetrahedra_points1, tetrahedra_points2)

        contact_point = np.sum(
            self.contact_coms * self.contact_areas[:, np.newaxis],
            axis=0) / sum(self.contact_areas)

        details = {
            "contact_polygons": self.contact_polygons,
            "contact_polygon_triangles": self.contact_polygon_triangles,
            "contact_planes": self.contact_planes,
            "intersecting_tetrahedra1": intersecting_tetrahedra1,
            "intersecting_tetrahedra2": intersecting_tetrahedra2,
            "contact_coms": self.contact_coms,
            "contact_forces": self.contact_forces,
            "contact_areas": self.contact_areas,
            "pressures": self.pressures,
            "contact_point": contact_point
        }
        return details

    def _transform_to_world(
            self, frame2world, tetrahedra_points1, tetrahedra_points2):
        self.contact_polygons = [transform_points(frame2world, contact_polygon)
                                 for contact_polygon in self.contact_polygons]
        plane_points = (self.contact_planes[:, :3]
                        * self.contact_planes[:, 3, np.newaxis])
        plane_points = transform_points(frame2world, plane_points)
        plane_normals = transform_directions(
            frame2world, self.contact_planes[:, :3])
        plane_distances = np.sum(plane_points * plane_normals, axis=1)
        self.contact_planes = np.hstack((plane_normals,
                                         plane_distances.reshape(-1, 1)))
        self.contact_coms = transform_points(
            frame2world, self.contact_coms)
        self.contact_forces = transform_directions(
            frame2world, self.contact_forces)
        n_intersections = len(self.intersecting_tetrahedra1)
        intersecting_tetrahedra1 = tetrahedra_points1[np.asarray(
            self.intersecting_tetrahedra1, dtype=int)]
        intersecting_tetrahedra1 = transform_points(
            self.frame2world, intersecting_tetrahedra1.reshape(
                n_intersections * 4, 3)
        ).reshape(n_intersections, 4, 3)
        intersecting_tetrahedra2 = tetrahedra_points2[
            np.asarray(self.intersecting_tetrahedra2, dtype=int)]
        intersecting_tetrahedra2 = transform_points(
            self.frame2world,
            intersecting_tetrahedra2.reshape(n_intersections * 4, 3)
        ).reshape(n_intersections, 4, 3)
        return intersecting_tetrahedra1, intersecting_tetrahedra2



"""Utilities to work with URDF files."""
import re
import warnings
import numpy as np


def fast_transform_manager_initialization(tm, frames, base):
    """Initialize a transform manager with many frames.

    Parameters
    ----------
    tm : pytransform3d.transform_manager.TransformManager
        Transform manager.

    frames : list
        Names of frames that are defined with respect to base frame.

    base : Hashable
        Name of the base frame.
    """
    if base not in tm.nodes:
        tm.nodes.append(base)

    for frame in frames:
        tm.nodes.append(frame)

        transform_key = (frame, base)

        ij_index = len(tm.i)
        tm.i.append(tm.nodes.index(frame))
        tm.j.append(tm.nodes.index(base))
        tm.transform_to_ij_index[transform_key] = ij_index

        tm.transforms[transform_key] = np.eye(4)

    tm._recompute_shortest_path()


def self_collision_whitelists(tm):
    """Construct self-collision whitelist.

    Parameters
    ----------
    tm : pytransform3d.urdf.UrdfTransformManager
        Transform manager that has colliders.

    Returns
    -------
    whitelist : dict
        Whitelists for self-collision detection.
    """
    whitelist = {}
    link_info = LinkInfo(tm)
    for obj in tm.collision_objects:
        link_frame = link_info.link(obj.frame)
        parent_frame = link_info.parent_link(link_frame)
        child_frame = link_info.child_link(link_frame)
        collision_objects_link = link_info.collision_frames_attached_to_link(link_frame)
        collision_objects_parent = link_info.collision_frames_attached_to_link(parent_frame)
        collision_objects_child = link_info.collision_frames_attached_to_link(child_frame)
        whitelist[obj.frame] = (collision_objects_link
                                + collision_objects_parent
                                + collision_objects_child)
    return whitelist


class LinkInfo:
    """Collect information about links from a UrdfTransformManager.

    Parameters
    ----------
    tm : pytransform3d.urdf.UrdfTransformManager
        Transform manager.
    """
    def __init__(self, tm):
        self.tm = tm
        self.parent_links = {}
        self.child_links = {}
        for child, parent in tm.transforms:
            self.parent_links[child] = parent
            self.child_links[parent] = child
        # HACK uses naming convention from URDF parser to extract link
        #      of a collision object
        self.prog_match_link = re.compile(r"collision:(.*)\/.*")

    def link(self, frame):
        """Finds link frame for a given collision object frame."""
        result = self.prog_match_link.match(frame)
        if result is None:
            warnings.warn(
                f"Couldn't extract link of collision object at frame '{frame}'")
            return None
        else:
            link_frame = result.group(1)
            return link_frame

    def child_link(self, link_frame):
        """Get child of a link."""
        return self._connected_link(self.child_links, link_frame)

    def parent_link(self, link_frame):
        """Get parent of a link."""
        return self._connected_link(self.parent_links, link_frame)

    @staticmethod
    def _connected_link(relation_info, link_frame):
        return relation_info.get(link_frame, None)

    def collision_frames_attached_to_link(self, link_frame):
        """Get all collision frames attached to a link frame."""
        collision_frames = []
        # HACK uses naming convention from URDF parser to extract link
        #      of a collision object
        prog = re.compile(f"collision:{link_frame}" + r"\/.*")
        for node in self.tm.nodes:
            if prog.match(node):
                collision_frames.append(node)
        return collision_frames



#!/usr/bin/env python
from setuptools import setup, find_packages
import distance3d


if __name__ == "__main__":
    with open("README.md", "r") as f:
        long_description = f.read()
    setup(name="distance3d",
          version=distance3d.__version__,
          author="Alexander Fabisch",
          author_email="afabisch@googlemail.com",
          url="https://github.com/AlexanderFabisch/distance3d",
          description="Distance computation and collision detection in 3D.",
          long_description=long_description,
          long_description_content_type="text/markdown",
          classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: BSD License",
              "Operating System :: OS Independent",
              "Topic :: Scientific/Engineering :: Mathematics",
              "Topic :: Scientific/Engineering :: Visualization",
          ],
          license="BSD-3-Clause",
          packages=find_packages(),
          install_requires=["numpy>=1.18", "scipy", "matplotlib",
                            "pytransform3d", "open3d", "aabbtree", "numba"],
          extras_require={
              "test": ["pytest", "pytest-cov"],
              "doc": ["sphinx", "sphinx-gallery", "numpydoc"],
          })



import math
import numba
import numpy as np


def point_to_line(point, line_point, line_direction):
    """Compute the shortest distance between point and line.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    Returns
    -------
    distance : float
        The shortest distance between point and line.

    closest_point_line : array, shape (3,)
        Closest point on line.
    """
    return _point_to_line(point, line_point, line_direction)[:2]


@numba.njit(
    numba.types.Tuple((numba.float64, numba.float64[::1], numba.float64))
    (numba.float64[::1], numba.float64[::1], numba.float64[::1]),
    cache=True)
def _point_to_line(point, line_point, line_direction):
    diff = point - line_point
    t = np.dot(line_direction, diff)
    direction_fraction = t * line_direction
    diff -= direction_fraction
    closest_point_line = line_point + direction_fraction
    return np.linalg.norm(diff), closest_point_line, t


@numba.njit(
    numba.types.Tuple((numba.float64, numba.float64[::1]))
    (numba.float64[::1], numba.float64[::1], numba.float64[::1]),
    cache=True)
def point_to_line_segment(point, segment_start, segment_end):
    """Compute the shortest distance between point and line segment.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    Returns
    -------
    distance : float
        The shortest distance between point and line segment.

    closest_point_line_segment : array, shape (3,)
        Closest point on line segment.
    """
    segment_direction = segment_end - segment_start
    # Project point onto segment, computing parameterized position
    # s(t) = segment_start + t * (segment_end - segment_start)
    t = (np.dot(point - segment_start, segment_direction) /
         np.dot(segment_direction, segment_direction))
    # If outside segment, clamp t to the closest endpoint
    t = min(max(t, 0.0), 1.0)
    # Compute projected position from the clamped t
    closest_point_line_segment = segment_start + t * segment_direction
    return (np.linalg.norm(point - closest_point_line_segment),
            closest_point_line_segment)


def line_to_line(line_point1, line_direction1, line_point2, line_direction2,
                 epsilon=1e-6):
    """Compute the shortest distance between two lines.

    Parameters
    ----------
    line_point1 : array, shape (3,)
        Point on the first line.

    line_direction1 : array, shape (3,)
        Direction of the first line. This is assumed to be of unit length.

    line_point2 : array, shape (3,)
        Point on the second line.

    line_direction2 : array, shape (3,)
        Direction of the second line. This is assumed to be of unit length.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    distance : float
        The shortest distance between two lines.

    closest_point_line1 : array, shape (3,)
        Closest point on first line.

    closest_point_line2 : array, shape (3,)
        Closest point on second line.
    """
    return _line_to_line(
        line_point1, line_direction1, line_point2, line_direction2,
        epsilon)[:3]


@numba.njit(
    numba.types.Tuple((numba.float64, numba.float64[::1], numba.float64[::1],
                       numba.float64, numba.float64))
    (numba.float64[::1], numba.float64[::1], numba.float64[::1],
     numba.float64[::1], numba.float64),
    cache=True)
def _line_to_line(line_point1, line_direction1, line_point2, line_direction2,
                  epsilon):
    diff = line_point1 - line_point2
    a12 = -np.dot(line_direction1, line_direction2)
    b1 = np.dot(line_direction1, diff)
    c = np.dot(diff, diff)
    det = 1.0 - a12 * a12

    if abs(det) >= epsilon:
        b2 = -np.dot(line_direction2, diff)
        t1 = (a12 * b2 - b1) / det
        t2 = (a12 * b1 - b2) / det
        dist_squared = (
            t1 * (t1 + a12 * t2 + 2.0 * b1)
            + t2 * (a12 * t1 + t2 + 2.0 * b2) + c)
        closest_point_line2 = line_point2 + t2 * line_direction2
    else:  # parallel lines
        t1 = -b1
        t2 = 0.0
        dist_squared = b1 * t1 + c
        closest_point_line2 = line_point2

    closest_point_line1 = line_point1 + t1 * line_direction1

    return (math.sqrt(abs(dist_squared)), closest_point_line1,
            closest_point_line2, t1, t2)


def line_to_line_segment(
        line_point, line_direction, segment_start, segment_end, epsilon=1e-6):
    """Compute the shortest distance between line and line segment.

    Implementation adapted from Real-Time Collision Detection by Christer
    Ericson published by Morgan Kaufmann Publishers, Copyright 2005 Elsevier
    Inc.

    Parameters
    ----------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line and line segment.

    closest_point1 : array, shape (3,)
        Closest point on line.

    closest_point2 : array, shape (3,)
        Closest point on line segment.
    """
    return _line_to_line_segment(
        line_point, line_direction, segment_start, segment_end, epsilon)[:3]


# modified version of line segment to line segment
@numba.njit(numba.types.Tuple(
    (numba.float64, numba.float64[::1], numba.float64[::1], numba.float64, numba.float64))
    (numba.float64[::1], numba.float64[::1], numba.float64[::1],
     numba.float64[::1], numba.float64), cache=True)
def _line_to_line_segment(
        line_point, line_direction, segment_start, segment_end, epsilon):
    # Segment direction vectors
    d = segment_end - segment_start

    # Squared segment lengths, always nonnegative
    a = np.dot(d, d)
    e = np.dot(line_direction, line_direction)

    if a < epsilon and e < epsilon:
        # Both segments degenerate into points
        return (np.linalg.norm(line_point - segment_start),
                segment_start, line_point, 0.0, 0.0)

    r = segment_start - line_point
    f = np.dot(line_direction, r)

    if a < epsilon:
        # First segment degenerates into a point
        s = 0.0
        t = f / e
    else:
        c = np.dot(d, r)
        if e <= epsilon:
            # Second segment degenerates into a point
            t = 0.0
            s = min(max(-c / a, 0.0), 1.0)
        else:
            # General nondegenerate case
            b = np.dot(d, line_direction)
            denom = a * e - b * b  # always nonnegative

            if denom != 0.0:
                # If segements not parallel, compute closest point on line 1 to
                # line 2 and clamp to segment 1.
                s = min(max((b * f - c * e) / denom, 0.0), 1.0)
            else:
                # Parallel case: compute arbitrary s.
                s = 0.0

            t = (b * s + f) / e

    closest_point1 = line_point + t * line_direction
    closest_point2 = segment_start + s * d

    return np.linalg.norm(closest_point2 - closest_point1), closest_point1, closest_point2, t, s


def line_segment_to_line_segment(
        segment_start1, segment_end1, segment_start2, segment_end2, epsilon=1e-6):
    """Compute the shortest distance between two line segments.

    Implementation adapted from Real-Time Collision Detection by Christer
    Ericson published by Morgan Kaufmann Publishers, Copyright 2005 Elsevier
    Inc.

    Parameters
    ----------
    segment_start1 : array, shape (3,)
        Start point of segment 1.

    segment_end1 : array, shape (3,)
        End point of segment 1.

    segment_start2 : array, shape (3,)
        Start point of segment 2.

    segment_end2 : array, shape (3,)
        End point of segment 2.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    distance : float
        The shortest distance between two line segments.

    closest_point_segment1 : array, shape (3,)
        Closest point on first line segment.

    closest_point_segment2 : array, shape (3,)
        Closest point on second line segment.
    """
    return _line_segment_to_line_segment(
        segment_start1, segment_end1, segment_start2, segment_end2, epsilon)


@numba.njit(numba.types.Tuple(
    (numba.float64, numba.float64[:], numba.float64[:]))
    (numba.float64[:], numba.float64[:], numba.float64[:],
     numba.float64[:], numba.float64), cache=True)
def _line_segment_to_line_segment(
        segment_start1, segment_end1, segment_start2, segment_end2, epsilon):
    # Segment direction vectors
    d1 = segment_end1 - segment_start1
    d2 = segment_end2 - segment_start2

    # Squared segment lengths, always nonnegative
    a = np.dot(d1, d1)
    e = np.dot(d2, d2)

    if a < epsilon and e < epsilon:
        # Both segments degenerate into points
        return (np.linalg.norm(segment_start2 - segment_start1),
                segment_start1, segment_start2)

    r = segment_start1 - segment_start2
    f = np.dot(d2, r)

    if a < epsilon:
        # First segment degenerates into a point
        s = 0.0
        t = min(max(f / e, 0.0), 1.0)
    else:
        c = np.dot(d1, r)
        if e <= epsilon:
            # Second segment degenerates into a point
            t = 0.0
            s = min(max(-c / a, 0.0), 1.0)
        else:
            # General nondegenerate case
            b = np.dot(d1, d2)
            denom = a * e - b * b  # always nonnegative

            if denom != 0.0:
                # If segements not parallel, compute closest point on line 1 to
                # line 2 and clamp to segment 1.
                s = min(max((b * f - c * e) / denom, 0.0), 1.0)
            else:
                # Parallel case: compute arbitrary s.
                s = 0.0

            t = (b * s + f) / e

            # If t in [0, 1] done. Else clamp t, recompute s.
            if t < 0.0:
                t = 0.0
                s = min(max(-c / a, 0.0), 1.0)
            elif t > 1.0:
                t = 1.0
                s = min(max((b - c) / a, 0.0), 1.0)

    closest_point1 = segment_start1 + s * d1
    closest_point2 = segment_start2 + t * d2

    return (np.linalg.norm(closest_point2 - closest_point1), closest_point1,
            closest_point2)



import numpy as np
import numba
from ..geometry import (
    hesse_normal_form, convert_segment_to_line, line_from_pluecker,
    convert_rectangle_to_vertices, convert_box_to_vertices,
    support_function_ellipsoid, support_function_cylinder)


def point_to_plane(point, plane_point, plane_normal, signed=False):
    """Compute the shortest distance between a point and a plane.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    signed : bool, optional (default: False)
        Should the distance have a sign?

    Returns
    -------
    dist : float
        The shortest distance between point and plane. A sign indicates
        the direction along the normal.

    closest_point_plane : array, shape (3,)
        Closest point on plane.
    """
    return _point_to_plane(point, plane_point, plane_normal, signed)


@numba.njit(
    numba.types.Tuple((numba.float64, numba.float64[::1]))
    (numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.boolean),
    cache=True)
def _point_to_plane(point, plane_point, plane_normal, signed):
    t = np.dot(plane_normal, point - plane_point)
    closest_point_plane = point - t * plane_normal
    if not signed:
        t = abs(t)
    return t, closest_point_plane


def line_to_plane(
        line_point, line_direction, plane_point, plane_normal, epsilon=1e-6):
    """Compute the shortest distance from line to plane.

    Parameters
    ----------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line and plane.

    closest_point_line : array, shape (3,)
        Closest point on line.

    closest_point_plane : array, shape (3,)
        Closest point on plane.
    """
    intersection, t = _line_to_plane(
        line_point, line_direction, plane_point, plane_normal, epsilon)

    if intersection:
        closest_point_line = line_point + t * line_direction
        return 0.0, closest_point_line, closest_point_line

    dist, closest_point_plane = point_to_plane(
        line_point, plane_point, plane_normal)
    return dist, line_point, closest_point_plane


@numba.njit(
    numba.types.Tuple((numba.boolean, numba.float64))
    (numba.float64[::1], numba.float64[::1], numba.float64[::1],
     numba.float64[::1], numba.float64),
    cache=True)
def _line_to_plane(line_point, line_direction, plane_point, plane_normal,
                   epsilon):
    length_of_dir_projected_on_normal = np.dot(line_direction, plane_normal)
    if length_of_dir_projected_on_normal * length_of_dir_projected_on_normal < epsilon:
        # line parallel to plane
        return False, 0.0

    _, d = hesse_normal_form(plane_point, plane_normal)
    t = (d - np.dot(plane_normal, line_point)) / np.dot(plane_normal, line_direction)
    return True, t


def line_segment_to_plane(
        segment_start, segment_end, plane_point, plane_normal, epsilon=1e-6):
    """Compute the shortest distance between line segment and plane.

    Parameters
    ----------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line segment and plane.

    closest_point_segment : array, shape (3,)
        Closest point on line segment.

    closest_point_plane : array, shape (3,)
        Closest point on plane.
    """
    return _line_segment_to_plane(
        segment_start, segment_end, plane_point, plane_normal, epsilon)


@numba.njit(cache=True)
def _line_segment_to_plane(
        segment_start, segment_end, plane_point, plane_normal, epsilon):
    segment_direction, segment_length = convert_segment_to_line(
        segment_start, segment_end)
    line_intersects, t = _line_to_plane(
        segment_start, segment_direction, plane_point, plane_normal, epsilon)

    if line_intersects:
        if 0 <= t <= segment_length:
            closest_point_segment = segment_start + t * segment_direction
            return 0.0, closest_point_segment, closest_point_segment

        if t < 0.0:
            closest_point_segment = segment_start
        else:
            closest_point_segment = segment_end
    else:
        closest_point_segment = segment_start

    dist, closest_point_plane = _point_to_plane(
        closest_point_segment, plane_point, plane_normal, False)
    return dist, closest_point_segment, closest_point_plane


def plane_to_plane(plane_point1, plane_normal1, plane_point2, plane_normal2,
                   epsilon=1e-6):
    """Compute the shortest distance between two planes.

    Parameters
    ----------
    plane_point1 : array, shape (3,)
        Point on the plane.

    plane_normal1 : array, shape (3,)
        Normal of the plane. We assume unit length.

    plane_point2 : array, shape (3,)
        Point on the plane.

    plane_normal2 : array, shape (3,)
        Normal of the plane. We assume unit length.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line segment and plane.

    closest_point_plane1 : array, shape (3,)
        Closest point on first plane.

    closest_point_plane2 : array, shape (3,)
        Closest point on second plane.
    """
    line_direction = np.cross(plane_normal1, plane_normal2)
    if np.linalg.norm(line_direction) > epsilon:
        line_direction, line_moment = plane_intersects_plane(
            plane_point1, plane_normal1, plane_point2, plane_normal2,
            line_direction=line_direction)
        line_point, _ = line_from_pluecker(line_direction, line_moment)
        return 0.0, line_point, line_point
    else:
        dist, closest_point_plane2 = point_to_plane(
            plane_point1, plane_point2, plane_normal2)
        return dist, plane_point1, closest_point_plane2


def plane_intersects_plane(
        plane_point1, plane_normal1, plane_point2, plane_normal2,
        line_direction=None):
    _, d1 = hesse_normal_form(plane_point1, plane_normal1)
    _, d2 = hesse_normal_form(plane_point2, plane_normal2)
    if line_direction is None:
        line_direction = np.cross(plane_normal1, plane_normal2)
    line_moment = plane_normal1 * d2 - plane_normal2 * d1
    return line_direction, line_moment


def plane_to_triangle(plane_point, plane_normal, triangle_points):
    """Compute the shortest distance between a plane and a triangle.

    Parameters
    ----------
    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    triangle_points : array, shape (3, 3)
        Each row contains a point of the triangle (A, B, C).

    Returns
    -------
    dist : float
        The shortest distance between triangle and plane.

    closest_point_plane : array, shape (3,)
        Closest point on plane.

    closest_point_triangle : array, shape (3,)
        Closest point on triangle.
    """
    return _plane_to_convex_hull_points(plane_point, plane_normal, triangle_points)


@numba.njit(cache=True)
def _plane_to_convex_hull_points(plane_point, plane_normal, points):
    ts = np.dot(points - plane_point.reshape(1, -1), plane_normal)
    min_idx = np.argmin(ts)
    max_idx = np.argmax(ts)

    if ts[min_idx] * ts[max_idx] < 0:  # on opposite sides, intersection
        return _line_segment_to_plane(
            points[min_idx], points[max_idx], plane_point, plane_normal, 1e-6)

    closest_idx = np.argmin(np.abs(ts))
    closest_point = points[closest_idx]
    t = ts[closest_idx]
    closest_point_plane = closest_point - t * plane_normal

    return abs(t), closest_point_plane, closest_point


def plane_to_rectangle(
        plane_point, plane_normal, rectangle_center, rectangle_axes,
        rectangle_lengths):
    """Compute the shortest distance between a plane and a rectangle.

    Parameters
    ----------
    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    Returns
    -------
    dist : float
        The shortest distance between rectangle and plane.

    closest_point_plane : array, shape (3,)
        Closest point on plane.

    closest_point_rectangle : array, shape (3,)
        Closest point on rectangle.
    """
    points = convert_rectangle_to_vertices(
        rectangle_center, rectangle_axes, rectangle_lengths)
    return _plane_to_convex_hull_points(plane_point, plane_normal, points)


def plane_to_box(plane_point, plane_normal, box2origin, size):
    """Compute the shortest distance between a plane and a box.

    Parameters
    ----------
    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box along its axes.

    Returns
    -------
    dist : float
        The shortest distance between rectangle and plane.

    closest_point_plane : array, shape (3,)
        Closest point on plane.

    closest_point_box : array, shape (3,)
        Closest point on box.
    """
    points = convert_box_to_vertices(box2origin, size)
    return _plane_to_convex_hull_points(plane_point, plane_normal, points)


@numba.njit(cache=True)
def plane_to_ellipsoid(plane_point, plane_normal, ellipsoid2origin, radii):
    """Compute the shortest distance between a plane and an ellipsoid.

    Parameters
    ----------
    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    ellipsoid2origin : array, shape (4, 4)
        Pose of the ellipsoid.

    radii : array, shape (3,)
        Radii of the ellipsoid.

    Returns
    -------
    dist : float
        The shortest distance between rectangle and plane.

    closest_point_plane : array, shape (3,)
        Closest point on plane.

    closest_point_ellipsoid : array, shape (3,)
        Closest point on ellipsoid.
    """
    point1 = support_function_ellipsoid(-plane_normal, ellipsoid2origin, radii)
    point2 = support_function_ellipsoid(plane_normal, ellipsoid2origin, radii)
    return _plane_to_convex_hull_points(
        plane_point, plane_normal, np.vstack((point1, point2)))


@numba.njit(cache=True)
def plane_to_cylinder(
        plane_point, plane_normal, cylinder2origin, radius, length):
    """Compute the shortest distance between a plane and a cylinder.

    Parameters
    ----------
    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    Returns
    -------
    dist : float
        The shortest distance between rectangle and plane.

    closest_point_plane : array, shape (3,)
        Closest point on plane.

    closest_point_cylinder : array, shape (3,)
        Closest point on cylinder.
    """
    point1 = support_function_cylinder(
        -plane_normal, cylinder2origin, radius, length)
    point2 = support_function_cylinder(
        plane_normal, cylinder2origin, radius, length)
    return _plane_to_convex_hull_points(
        plane_point, plane_normal, np.vstack((point1, point2)))



"""Self-collision detection."""
from . import gjk


def detect(bvh):
    """Detect self collisions of a robot represented by a BVH.

    Parameters
    ----------
    bvh : BoundingVolumeHierarchy
        Bounding volume hierarchy that contains colliders of a robot. Note that
        the attribute bvh.self_collision_whitelists_ has to be filled.
        Otherwise a collection of a collider with itself or direct neighbors
        will be considered a self collision!

    Returns
    -------
    contacts : dict
        Maps each collider frame to a boolean indicating whether it is in
        contact with another collider or not.
    """
    contacts = {}
    for frame, collider in bvh.colliders_.items():
        if frame in contacts:
            continue  # contact was detected before

        candidates = bvh.aabb_overlapping_colliders(
            collider, whitelist=bvh.self_collision_whitelists_[frame])

        contacts[frame] = False
        for frame2, collider2 in candidates.items():
            if gjk.gjk_intersection(collider, collider2):
                contacts[frame] = True
                contacts[frame2] = True
                break
    return contacts


def detect_any(bvh):
    """Detect self collisions of a robot represented by a BVH.

    This function aborts on first detected self collision.

    Parameters
    ----------
    bvh : BoundingVolumeHierarchy
        Bounding volume hierarchy that contains colliders of a robot. Note that
        the attribute bvh.self_collision_whitelists_ has to be filled.
        Otherwise a collection of a collider with itself or direct neighbors
        will be considered a self collision!

    Returns
    -------
    has_self_collision : bool
        Whether there is any self collision or not.
    """
    for frame, collider in bvh.colliders_.items():
        candidates = bvh.aabb_overlapping_colliders(
            collider, whitelist=bvh.self_collision_whitelists_[frame])

        for frame2, collider2 in candidates.items():
            if gjk.gjk_intersection(collider, collider2):
                return True
    return False



import numpy as np
from ..utils import transform_points, invert_transform
from ._tetra_mesh_creation import (
    make_tetrahedral_sphere, make_tetrahedral_ellipsoid,
    make_tetrahedral_cube, make_tetrahedral_box, make_tetrahedral_cylinder,
    make_tetrahedral_capsule)
from ._mesh_processing import center_of_mass_tetrahedral_mesh, tetrahedral_mesh_aabbs
from ..aabb_tree import AabbTree
from ..visualization import RigidBodyTetrahedralMesh


class RigidBody:
    """Rigid body represented by tetrahedral mesh.

    Parameters
    ----------
    body2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    potentials : array, shape (n_vertices,)
        Potential of each vertex. Shortest distance to surface.

    youngs_modulus : float, optional (default: 1.0)
        The stiffness of the material the rigidbody is made out of.
    """
    def __init__(self, body2origin, vertices, tetrahedra, potentials,
                 youngs_modulus=1.0):
        self.body2origin_ = body2origin
        self.vertices_ = vertices
        self.tetrahedra_ = tetrahedra
        self.potentials_ = potentials

        self._tetrahedra_points = None
        self._com = None

        self._aabbs = None
        self._aabb_tree = None

        self._youngs_modulus = youngs_modulus
        self._artist = None

    @staticmethod
    def make_sphere(center, radius, order=4):
        """Create sphere.

        Parameters
        ----------
        center : array, shape (3,)
            Center of the sphere.

        radius : float
            Radius of the sphere.

        order : int, optional (default: 4)
            Number of subdivisions of initial 20 triangles.

        Returns
        -------
        rigid_body : RigidBody
            Sphere.
        """
        mesh2origin = np.eye(4)
        mesh2origin[:3, 3] = center
        vertices, tetrahedra, potentials = make_tetrahedral_sphere(
            radius, order)
        return RigidBody(mesh2origin, vertices, tetrahedra, potentials)

    @staticmethod
    def make_ellipsoid(ellipsoid2origin, radii, order=4):
        """Create ellipsoid.

        Parameters
        ----------
        ellipsoid2origin : array, shape (4, 4)
            Pose of the ellipsoid.

        radii : array, shape (3,)
            Radii of the ellipsoid.

        order : int, optional (default: 4)
            Number of subdivisions of initial 20 triangles.

        Returns
        -------
        rigid_body : RigidBody
            Ellipsoid.
        """
        vertices, tetrahedra, potentials = make_tetrahedral_ellipsoid(
            radii, order)
        return RigidBody(ellipsoid2origin, vertices, tetrahedra, potentials)

    @staticmethod
    def make_cube(cube2origin, size):
        """Create cube.

        Parameters
        ----------
        cube2origin : array, shape (4, 4)
            Pose of the cube.

        size : float
            Length of the edges in each dimension.

        Returns
        -------
        rigid_body : RigidBody
            Cube.
        """
        vertices, tetrahedra, potentials = make_tetrahedral_cube(size)
        return RigidBody(cube2origin, vertices, tetrahedra, potentials)

    @staticmethod
    def make_box(box2origin, size):
        """Create box.

        Parameters
        ----------
        box2origin : array, shape (4, 4)
            Pose of the box.

        size : array, shape (3,)
            Lengths of the edges in each dimension.

        Returns
        -------
        rigid_body : RigidBody
            Box.
        """
        vertices, tetrahedra, potentials = make_tetrahedral_box(size)
        return RigidBody(box2origin, vertices, tetrahedra, potentials)

    @staticmethod
    def make_cylinder(cylinder2origin, radius, length, resolution_hint=0.1):
        """Create cylinder.

        Parameters
        ----------
        cylinder2origin : array, shape (4, 4)
            Pose of the cylinder.

        radius : float
            Radius of the cylinder.

        length : float
            Length of the cylinder.

        resolution_hint : float, optional (default: 0.1)
            Controls the fineness of the tetrahedral mesh. The coarsest mesh
            that produces desirable results will allow simulation to run as
            efficiently as possible. The circles of the cylinder will have
            2 * pi * radius / resolution_hint edges.

        Returns
        -------
        rigid_body : RigidBody
            Cylinder.
        """
        vertices, tetrahedra, potentials = make_tetrahedral_cylinder(
            radius, length, resolution_hint)
        return RigidBody(cylinder2origin, vertices, tetrahedra, potentials)

    @staticmethod
    def make_capsule(capsule2origin, radius, height, resolution_hint=0.1):
        """Create capsule.

        Parameters
        ----------
        capsule2origin : array, shape (4, 4)
            Pose of the capsule.

        radius : float
            Radius of the capsule.

        height : float
            Height of the capsule.

        resolution_hint : float, optional (default: 0.1)
            Controls the fineness of the tetrahedral mesh. The coarsest mesh
            that produces desirable results will allow simulation to run as
            efficiently as possible. The circles of the cylinder and great
            circles of each hemisphere will have 2 * pi * radius /
            resolution_hint edges.

        Returns
        -------
        rigid_body : RigidBody
            Capsule.
        """
        vertices, tetrahedra, potentials = make_tetrahedral_capsule(
            radius, height, resolution_hint)
        return RigidBody(capsule2origin, vertices, tetrahedra, potentials)

    @property
    def tetrahedra_points(self):
        """All points of all tetrahedra in body frame."""
        if self._tetrahedra_points is None:
            self._tetrahedra_points = self.vertices_[self.tetrahedra_]
        return self._tetrahedra_points

    @property
    def tetrahedra_potentials(self):
        """Potentials of tetrahedra."""
        return self.potentials_[self.tetrahedra_]

    @property
    def com(self):
        """Center of mass."""
        if self._com is None:
            self._com = center_of_mass_tetrahedral_mesh(self.tetrahedra_points)
        return self._com

    def express_in(self, new_body2origin):
        """Express tetrahedral meshes in another frame.

        Vertices will be transformed so that they are expressed in the new
        frame.

        Parameters
        ----------
        new_body2origin : array, shape (4, 4)
            New frame.
        """
        origin2new_body = invert_transform(new_body2origin)
        body2new_body = np.dot(origin2new_body, self.body2origin_)
        self.vertices_ = transform_points(body2new_body, self.vertices_)
        self.body2origin_ = np.copy(new_body2origin)

        self._tetrahedra_points = None
        self._com = None
        self._aabbs = None
        self._aabb_tree = None

    def update_pose(self, body2origin):
        """Set new pose.

        Parameters
        ----------
        body2origin : array, shape (4, 4)
            New pose.
        """
        # TODO Check if this is correct.
        self.body2origin_ = body2origin
        if self.artist_ is not None:
            self.artist_.set_data(body2origin, vertices=self.vertices_, tetrahedra=self.tetrahedra_)

    def aabb(self):
        """The aabb of the rigidbody"""
        return self.aabb_tree.get_root_aabb()

    @property
    def aabbs(self):
        """The aabbs for broad phase collision detection"""
        if self._aabbs is None:
            self._aabbs = tetrahedral_mesh_aabbs(self.tetrahedra_points)
        return self._aabbs

    @property
    def aabb_tree(self):
        """The aabb_tree for broad phase collision detection"""
        if self._aabb_tree is None:
            self._aabb_tree = AabbTree()
            self._aabb_tree.insert_aabbs(self.aabbs, pre_insertion_methode="sort")
        return self._aabb_tree

    @property
    def youngs_modulus(self):
        """Get the young's modulus of the Object. (stiffness)"""
        return self._youngs_modulus

    @youngs_modulus.setter
    def youngs_modulus(self, value):
        """Set the young's modulus of the Object. (stiffness)"""
        self._youngs_modulus = value

    @property
    def artist_(self):
        """The artist to visualize the rigid body as a wireframe."""
        if self._artist is None:
            self.make_artist()
        return self._artist

    def make_artist(self, c=None):
        """Creates a new artist.

            Parameters
            ----------
            c : array, shape (3
                The color of the wireframe.
        """
        self._artist = RigidBodyTetrahedralMesh(self.body2origin_, self.vertices_, self.tetrahedra_, c)
